/* Name: smpusb.c
 * Project: SMP reader device for Electronika MK-90
 * Creation Date: 2009-07-31
 * Copyright: (c) 2009 by Ilya Danilov
 * License: GNU GPL v2 (see License.txt), GNU GPL v3 or proprietary (CommercialLicense.txt)
 * This Revision: $Id: set-led.c 692 2008-11-07 15:07:40Z cs $
 */

/*
General Description:

This is a host control program for SMP USB-adapter. 

Exit codes are:
     0 - operation OK
     1 - unknown command or command absent
     2 - missing filename
     3 - file operation failed (read or write)
     4 - verifying failed (after writing)
     10 - USB device not found

This program must be linked with libusb on Unix and libusb-win32 on Windows.
See http://libusb.sourceforge.net/ or http://libusb-win32.sourceforge.net/
respectively.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <usb.h>        /* this is libusb */
#include "opendevice.h" /* common code moved to separate module */

#include "../requests.h"   /* custom request numbers */
#include "../usbconfig.h"  /* device's VID/PID and names */

#define strcasecmp  strcmp

#define CARDSIZE    10240
#define MAXCARDSIZE 65536
#define CHUNK_SIZE  4096

unsigned char       debug = 0;

int get_opt_key (int argc, char **argv, char *key) {
  int i;
  for (i=0;i<argc;i++) {
    if (!strcmp(key,argv[i])) return 1; 
  }
  return 0;
}

int get_opt_int (int argc, char **argv, char *key) {
  int i;
  for (i=0;i<argc;i++) {
    if (!strcmp(key,argv[i]) && i!=(argc-1)) {
      return atoi(argv[i+1]);
    }
  }
  return -1;
}

char *get_opt_string(int argc, char **argv, char *key) {
  int i;
  for (i=0;i<argc;i++) {
    if (!strcmp(key,argv[i]) && i!=(argc-1)) return argv[i+1]; 
  }
  return NULL;
}
static void usage()
{
    printf("\nusage:\n\n  smpusb.exe [command] [-f filename] [options]");
    printf("\n\nwhere command one of:\n\n");
    printf("  read  ..... perform cartridge read to filename\n"); 
    printf("  write ..... perform cartridge write and verify from filename\n"); 
    printf("  voltage  .. show internal battery voltage\n"); 
    printf("\n\noptions maybe:\n\n");
    printf("  -s NNN .... override SMP card size (default = %d)\n",CARDSIZE);
    printf("  -c NNN .... redefine USB chunk size (default = %d)\n",CHUNK_SIZE);
    printf("  -p  ....... disable low voltage warning (default = yes)\n");
    printf("  -d  ....... add debug output\n");
}

int get_voltage(usb_dev_handle *handle, float *voltage) {
  int cnt;
  unsigned short buffer[2];
  
  cnt = usb_control_msg(handle, 
                        USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
                        CUSTOM_RQ_GET_VOLTAGE, 0, 0, 
                        (char *)buffer, sizeof(buffer), 5000);
  if(cnt < sizeof(buffer)) {
    if(cnt < 0){
       printf("USB error: %s\n", usb_strerror());
    }else{
       printf("Internal error: only %d bytes received.\n", cnt);
    }
    return -1;
  }else{
    if (debug) printf("Read ADC values: [0] = %d, [1] = %d\n",buffer[0],buffer[1]);
    if (!buffer[1]) 
      *voltage = 0; // prevent zero division fault
    else
      *voltage = buffer[0]*1.235/buffer[1];
    return 0;
  }
}

int compare_buffers(unsigned char *read_buffer, unsigned char *verify_buffer, unsigned int buf_len) {
  unsigned int i;
  int error_flag = 0;
  for (i=0;i<buf_len;i++) {
    if (read_buffer[i] != verify_buffer[i]) {
      if (!error_flag) { 
        error_flag = 1;
        printf("Verification ERROR!\nThere are differences between source after control reading\n");
        printf("==========\n");
      }
      printf("%X: %02x %02x\n",i,read_buffer[i],verify_buffer[i]);
    } 
  }
  if (error_flag) {
        printf("==========\n");
        printf("WARNING: write operation FAILED!\n");
  } else 
    printf("Verification SMP contents after writing: OK\n");

  return error_flag;
}

int output_buffer(unsigned char *read_buffer, unsigned int buf_len, char *filename) {
  FILE *dst = fopen(filename,"wb");
  if ( dst != NULL ){
    fwrite(read_buffer,buf_len,1,dst);
    fclose(dst);
  } else {
    perror("Error opening file: "); 
    return 0;
  }
  return 1;
}

int input_buffer(unsigned char *read_buffer, unsigned int buf_len, char *filename) {
  FILE *src = fopen(filename,"rb");
  if ( src != NULL ){
    fread(read_buffer,buf_len,1,src);
    fclose(src);
  } else {
    perror("Error opening file: "); 
    return 0;
  }
  return 1;
}

int main(int argc, char **argv)
{
usb_dev_handle      *handle = NULL;
const unsigned char rawVid[2] = {USB_CFG_VENDOR_ID}, rawPid[2] = {USB_CFG_DEVICE_ID};
char                vendor[] = {USB_CFG_VENDOR_NAME, 0}, product[] = {USB_CFG_DEVICE_NAME, 0};
unsigned char       read_buffer[MAXCARDSIZE];
unsigned char       verify_buffer[MAXCARDSIZE];
int                 cnt, vid, pid;
float               voltage;

char                *filename = NULL;
unsigned char       voltage_warning = 1;
unsigned int        chunk_size = CHUNK_SIZE;
unsigned int        card_size = CARDSIZE;

    usb_init();
    if(argc < 2){   
        usage();
        exit(1);
    }

    /* parse command line */
    filename = get_opt_string(argc,argv,"-f");
    if (get_opt_key(argc,argv,"-p")) voltage_warning = 0;
    if (get_opt_key(argc,argv,"-d")) debug = 1;
    if (get_opt_key(argc,argv,"-c")) { 
      int val = get_opt_int(argc,argv,"-c");
      if (val == -1) val = CHUNK_SIZE;
      if ( (val<8) || (val>4096) ) {
        if (debug) {
          printf("Illegal USB chunk size %d. Default value %d wil be used\n",val,CHUNK_SIZE);
          val = CHUNK_SIZE;
        }
      }
      chunk_size = val;
    }
    if (get_opt_key(argc,argv,"-s")) {
      int val = get_opt_int(argc,argv,"-s");
      if (val == -1) val = CARDSIZE;
      if ( (val<1) || (val>MAXCARDSIZE) ) {
        if (debug) {
          printf("Illegal card size %d. Default value %d wil be used\n",val,CARDSIZE);
          val = CHUNK_SIZE;
        }
      }
      card_size = val;
    }

    /* compute VID/PID from usbconfig.h so that there is a central source of information */
    vid = rawVid[1] * 256 + rawVid[0];
    pid = rawPid[1] * 256 + rawPid[0];
    /* The following function is in opendevice.c: */
    if(usbOpenDevice(&handle, vid, vendor, pid, product, NULL, NULL, NULL, debug) != 0){
        printf("Could not find USB device \"%s\" with vid=0x%x pid=0x%x\n", product, vid, pid);
        exit(10);
    }
    /* Since we use only control endpoint 0, we don't need to choose a
     * configuration and interface. Reading device descriptor and setting a
     * configuration and interface is done through endpoint 0 after all.
     * However, newer versions of Linux require that we claim an interface
     * even for endpoint 0. Enable the following code if your operating system
     * needs it: */
#if 0
    int retries = 1, usbConfiguration = 1, usbInterface = 0;
    if(usb_set_configuration(handle, usbConfiguration) && showWarnings){
        fprintf(stderr, "Warning: could not set configuration: %s\n", usb_strerror());
    }
    /* now try to claim the interface and detach the kernel HID driver on
     * Linux and other operating systems which support the call. */
    while((len = usb_claim_interface(handle, usbInterface)) != 0 && retries-- > 0){
#ifdef LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP
        if(usb_detach_kernel_driver_np(handle, 0) < 0 && showWarnings){
            fprintf(stderr, "Warning: could not detach kernel driver: %s\n", usb_strerror());
        }
#endif
    }
#endif

    if(strcasecmp(argv[1], "voltage") == 0){
        if (!get_voltage(handle,&voltage)) {
          printf("Battery voltage is %.2f V\n",voltage);
        }
    } else if(strcasecmp(argv[1], "read") == 0) {
        unsigned int remaining_bytes = card_size;
        unsigned int buffer_pointer = 0;
        if (!filename) {
          printf("Error: file name needed for this operation.\n");
          usage();
          exit(2);
        }
        while (remaining_bytes) {
          cnt = usb_control_msg( handle, 
                                 USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
                                 CUSTOM_RQ_READ_DATA, buffer_pointer, 0, 
                                 read_buffer+buffer_pointer, 
                                 ((chunk_size <= remaining_bytes) ? chunk_size : remaining_bytes), 
                                 5000); 
          if (cnt < 0) {
            printf("USB error: %s\n", usb_strerror());
            exit(10);
          } else {
            if (debug) printf("..read chunk from %d (%d)\n", buffer_pointer, ((chunk_size <= remaining_bytes) ? chunk_size : remaining_bytes) );
            remaining_bytes -= cnt;
            buffer_pointer += cnt;
          }
        }
        if (!output_buffer(read_buffer,card_size,filename)) {
          exit(3);
        }
        if (voltage_warning && !get_voltage(handle,&voltage)) {
          if (voltage < 2.6) {
            printf("\nWARNING! Battery voltage is %.2f V, that is lower than recommended for storage ( 2.6 V )\nPossibly, data you read is a junk\n",voltage);
          } else if (voltage < 2.7) {
            printf("\nWARNING! Battery voltage is %.2f V, don'forget to replace battery before deadline ( 2.6 V )\n",voltage);
          }
        }
    } else if (strcasecmp(argv[1], "write") == 0) {
        unsigned int remaining_bytes = card_size;
        unsigned int buffer_pointer = 0;
        if (!filename) {
          printf("Error: file name needed for this operation.\n");
          usage();
          exit(2);
        }
        if (!input_buffer(read_buffer,card_size,filename)) {
          exit(3);
        }
        while (remaining_bytes) {
          cnt = usb_control_msg( handle, 
                                 USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT, 
                                 CUSTOM_RQ_WRITE_DATA, buffer_pointer, 0, 
                                 read_buffer+buffer_pointer, 
                                 ((chunk_size <= remaining_bytes) ? chunk_size : remaining_bytes), 
                                 5000); 
          if (cnt < 0) {
            printf("USB error: %s\n", usb_strerror());
            exit(10);
          } else {
            if (debug) printf("..write chunk from %d (%d)\n", buffer_pointer, ((chunk_size <= remaining_bytes) ? chunk_size : remaining_bytes) );
            remaining_bytes -= cnt;
            buffer_pointer += cnt;
          }
        }
        /* read written data */
        remaining_bytes = card_size;
        buffer_pointer = 0;
        while (remaining_bytes) {
          cnt = usb_control_msg( handle, 
                                 USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
                                 CUSTOM_RQ_READ_DATA, buffer_pointer, 0, 
                                 verify_buffer+buffer_pointer, 
                                 ((chunk_size <= remaining_bytes) ? chunk_size : remaining_bytes), 
                                 5000); 
          if (cnt < 0) {
            printf("USB error: %s\n", usb_strerror());
            exit(10);
          } else {
            if (debug) printf("..read chunk from %d (%d)\n", buffer_pointer, ((chunk_size <= remaining_bytes) ? chunk_size : remaining_bytes) );
            remaining_bytes -= cnt;
            buffer_pointer += cnt;
          }
        }
        /* compare two arrays */
        if (compare_buffers(read_buffer,verify_buffer,card_size)) {
          exit(4);
        }
        if (voltage_warning && !get_voltage(handle,&voltage)) {
          if (voltage < 2.6) {
            printf("\nWARNING! Battery voltage is %.2f V, that is lower than recommended for storage ( 2.6 V )\nPossibly, data you wrote will be destroyed after SMP will be disconnected from reader.\n",voltage);
          }else if (voltage < 2.7) {
            printf("\nWARNING! Battery voltage is %.2f V, don'forget to replace battery before deadline ( 2.6 V )\n",voltage);
          }
        }
    }else{
        usage(argv[0]);
        exit(1);
    }
    usb_close(handle);
    return 0;
}
